#include "protoGraph.h"

ProtoGraph::Edge::Edge(Vertice& a, Vertice& b)
 : term_a(a), prev_a(NULL), next_a(NULL),
   term_b(b), prev_b(NULL), next_b(NULL)
{
    AppendToVertices();
}

ProtoGraph::Edge::~Edge()
{
    DetachFromVertices();
}

void ProtoGraph::Edge::AppendToVertices()
{
    if ((prev_a = term_a.GetTailEdge()))
        prev_a->Append(term_a, this);
    else
        term_a.SetHeadEdge(this);
    term_a.SetTailEdge(this); 
    next_a = NULL;  
    
    if ((prev_b = term_b.GetTailEdge()))
        prev_b->Append(term_b, this);
    else
        term_b.SetHeadEdge(this);
    term_b.SetTailEdge(this); 
    next_b = NULL; 
}  // end ProtoGraph::Edge::AppendToVertice()

void ProtoGraph::Edge::DetachFromVertices()
{
   if (NULL != prev_a)
       prev_a->Append(term_a, next_a);
   else
       term_a.SetHeadEdge(next_a);
   if (NULL != next_a)
       next_a->Prepend(term_a, prev_a);
   else
       term_a.SetTailEdge(prev_a);
   prev_a = next_a = NULL; 
       
  if (NULL != prev_b)
      prev_b->Append(term_b, next_b);
   else
       term_b.SetHeadEdge(next_b);
   if (NULL != next_b)
       next_b->Prepend(term_b, prev_b);
   else
       term_b.SetTailEdge(prev_b); 
   prev_b = next_b = NULL; 
}  // end ProtoGraph::Edge::DetachVertice()
        
ProtoGraph::Vertice::Vertice()
 : edge_list_head(NULL), edge_list_tail(NULL), 
   //queue_state_head(NULL), 
   graph_prev(NULL), graph_next(NULL)
{
}

ProtoGraph::Vertice::~Vertice()
{
    // Make sure no queues still refer to us!
    //ASSERT(NULL == queue_state_head);
    ASSERT(queue_state_tree.IsEmpty());
    // Delete edges associated with this vertice
    Edge* next = edge_list_head;
    while (NULL != next)
    {
        Edge* edge = next;
        next = next->GetNext(*this);
        delete edge;      
    }
    edge_list_head = edge_list_tail = NULL;
} 

ProtoGraph::Vertice::Queue::State::State()
 : entry(*this), queue(NULL), vertice(NULL)
{
}

ProtoGraph::Vertice::Queue::State::~State()
{
    // (TBD) Should we remove from "queue" if it is non-NULL?
    if (NULL != vertice) vertice->Dereference(*this);
}

ProtoGraph::Vertice::Queue::State::Entry::Entry(State& theState)
 : ProtoTree::Item((const char*)(&(theState.queue)), sizeof(Queue*) << 3), queue_state(theState)
{
    const Queue** ptr = (const Queue**)key;
}


void ProtoGraph::Vertice::Reference(Queue::State& queueState)
{
    queueState.SetVertice(this);
    
    const Queue** ptr = (const Queue**)queueState.AccessEntry().GetKey();
    
    queue_state_tree.Insert(queueState.AccessEntry());
    
    /*queueState.PrependState(NULL);
    queueState.AppendState(queue_state_head);
    if (NULL != queue_state_head)
        queue_state_head->PrependState(&queueState);
    queue_state_head = &queueState;*/
}  // end ProtoGraph::Vertice::Reference()

void ProtoGraph::Vertice::Dereference(Queue::State& queueState)
{
    ASSERT(this == queueState.GetVertice());
    
    const Queue** ptr = (const Queue**)queueState.AccessEntry().GetKey();
    queue_state_tree.Remove(queueState.AccessEntry());
    /*Queue::State* prevState = queueState.GetStatePrev();
    Queue::State* nextState = queueState.GetStateNext();
    if (NULL != prevState)
        prevState->AppendState(nextState);
    else
        queue_state_head = nextState;
    if (NULL != nextState)
        nextState->PrependState(prevState);*/
    queueState.SetVertice(NULL);
}  // end ProtoGraph::Vertice::Dereference()

ProtoGraph::Vertice::Queue::State* ProtoGraph::Vertice::GetQueueState(const ProtoGraph::Vertice::Queue& queue) const
{
    const Queue* ptr = &queue;
    Queue::State::Entry* entry = static_cast<Queue::State::Entry*>(queue_state_tree.Find((const char*)&ptr, sizeof(Queue*) << 3));
    if (NULL != entry)
    {
        return &entry->GetQueueState();
    }
    else
    {
        /*Queue::State* nextState = queue_state_head;
        while (NULL != nextState)
        {
            if (&queue == nextState->GetQueue())
                return nextState;
            nextState = nextState->GetStateNext();   
        }*/
        return NULL;  // matching queue state not found
    }
}  // end ProtoGraph::Vertice::GetQueueState()

ProtoGraph::Vertice::EdgeIterator::EdgeIterator(const Vertice& theVertice)
 : vertice(theVertice), next_edge(theVertice.GetHeadEdge())
{
}

ProtoGraph::Vertice::EdgeIterator::~EdgeIterator()
{
}


ProtoGraph::Vertice::SimpleQueue::Item::Item()
 : prev(NULL), next(NULL)
{
}

ProtoGraph::Vertice::SimpleQueue::Item::~Item()
{
    // (TBD) Should we remove from "queue" if it is non-NULL?
}

ProtoGraph::Vertice::SimpleQueue::SimpleQueue(SimpleQueue::ItemPool* itemPool)
 : head(NULL), tail(NULL), item_pool(itemPool)
{
}

ProtoGraph::Vertice::SimpleQueue::~SimpleQueue()
{
}


bool ProtoGraph::Vertice::SimpleQueue::Prepend(Vertice& vertice)
{
    Item* item = GetNewItem();
    if (NULL == item)
    {
        DMSG(0, "ProtoGraph::Vertice::SimpleQueue::Prepend() NewItem() error: %s\n", GetErrorString());
        return false;
    }
    item->SetQueue(this);
    Associate(*item, vertice);
    PrependItem(*item);
    return true;
}  // end ProtoGraph::Vertice::SimpleQueue::Prepend()

bool ProtoGraph::Vertice::SimpleQueue::Append(Vertice& vertice)
{
    Item* item = GetNewItem();
    if (NULL == item)
    {
        DMSG(0, "ProtoGraph::Vertice::SimpleQueue::Append() NewItem() error: %s\n", GetErrorString());
        return false;
    }
    item->SetQueue(this);
    Associate(*item, vertice);
    AppendItem(*item);
    return true;
}  // end ProtoGraph::Vertice::SimpleQueue::Append()

void ProtoGraph::Vertice::SimpleQueue::Remove(Vertice& vertice)
{
    // 1) Retrieve "Item*" from "vertice" queue state
    Item* item = static_cast<Item*>(vertice.GetQueueState(*this));
    // 2) Remove "item" from our linked list 
    ASSERT(NULL != item);
    RemoveItem(*item);
    Disassociate(*item, vertice);
    // 3) Return "item" to pool
    if (NULL != item_pool)
        item_pool->Put(item);
    else
        delete item;
}  // end ProtoGraph::Vertice::SimpleQueue::Remove()

void ProtoGraph::Vertice::SimpleQueue::PrependItem(Item& item)
{
    item.Prepend(NULL);
    if (NULL == head)
        tail = &item;
    else
        head->Prepend(&item);
    item.Append(head);
    head = &item;
}  // end ProtoGraph::Vertice::SimpleQueue::PrependItem()

void ProtoGraph::Vertice::SimpleQueue::AppendItem(Item& item)
{
    item.Prepend(NULL);
    if (NULL == tail)
        head = &item;
    else
        tail->Append(&item);
    item.Prepend(tail);
    tail = &item;
}  // end ProtoGraph::Vertice::SimpleQueue::AppendItem()

void ProtoGraph::Vertice::SimpleQueue::RemoveItem(Item& item)
{
    Item* prevItem = item.GetPrev();
    Item* nextItem = item.GetNext();
    if (NULL != prevItem)
        prevItem->Append(nextItem);
    else
        head = nextItem;
    if (NULL != nextItem)
        nextItem->Prepend(prevItem);
    else
        tail = prevItem;
}  // end ProtoGraph::Vertice::SimpleQueue::RemoveItem()


ProtoGraph::Vertice::SimpleQueue::Iterator::Iterator(const SimpleQueue& theQueue, bool reverse)
 : queue(theQueue), next_item(reverse ? theQueue.tail : theQueue.head), forward(!reverse)
{
}

ProtoGraph::Vertice::SimpleQueue::Iterator::~Iterator()
{
}

void ProtoGraph::Vertice::SimpleQueue::Iterator::Reset()
{
    next_item = forward ? queue.head : queue.tail;
}  // end ProtoGraph::Vertice::SimpleQueue::Iterator::Reset()

ProtoGraph::Vertice* ProtoGraph::Vertice::SimpleQueue::Iterator::GetNextVertice()
{
    Item* item = next_item;
    next_item = item ? (forward ? item->GetNext() : item->GetPrev()) : NULL;
    return item ? item->GetVertice() : NULL;
}  // end ProtoGraph::Vertice::SimpleQueue::Iterator::GetNextVertice()


ProtoGraph::Vertice::SimpleQueue::ItemPool::ItemPool()
 : head(NULL)
{
}

ProtoGraph::Vertice::SimpleQueue::ItemPool::~ItemPool()
{
    Destroy();
}

void ProtoGraph::Vertice::SimpleQueue::ItemPool::Destroy()
{
    while (!IsEmpty())
    {
        Item* item = Get();
        delete item;
    }
}  // end ProtoGraph::Vertice::SimpleQueue::ItemPool::Destroy()

ProtoGraph::Vertice::SimpleQueue::Item* ProtoGraph::Vertice::SimpleQueue::ItemPool::Get()
{
    if (IsEmpty())
    {
        Item* item = new Item();
        if (NULL == item)
        {
            DMSG(0, "ProtoGraph::Vertice::SimpleQueue::ItemPool::Get() new Item error: %s\n",
                    GetErrorString());
            return NULL;
        }
        else
        {
            return item;
        }
    }
    else
    {
        Item* item = head;
        head = item->GetNext();
        return item;
    }
}  // end ProtoGraph::Vertice::SimpleQueue::ItemPool::Get()


ProtoGraph::ProtoGraph()
 : head_vertice(NULL), tail_vertice(NULL)
{
}

ProtoGraph::~ProtoGraph()
{
    //while (NULL != head_vertice)
    {
        //Vertice* vertice = head_vertice;
        // 1) Remove/dereference "vertice" from any queues (TBD!!)
        //TRACE("Protograph dtor, removing  vertice %p\n", vertice);
        //RemoveVertice(*vertice);   
        //TRACE("Protograph dtor, deleting  vertice %p\n", vertice);
        //delete vertice;
    }
}

void ProtoGraph::InsertVertice(Vertice& vertice)
{
    // Put new vertice at tail of linked list
    vertice.GraphPrepend(tail_vertice);
    vertice.GraphAppend(NULL);
    if (NULL != tail_vertice)
        tail_vertice->GraphAppend(&vertice);
    else
        head_vertice = &vertice;
    tail_vertice = &vertice;
}  // end ProtoGraph::InsertNode()

void ProtoGraph::RemoveVertice(Vertice& vertice)
{
    // Remove vertice from linked list
    Vertice* prevVertice = vertice.GetGraphPrev();
    Vertice* nextVertice = vertice.GetGraphNext();
    if (NULL != prevVertice)
        prevVertice->GraphAppend(nextVertice);
    else
        head_vertice = nextVertice;
    if (NULL != nextVertice)
        nextVertice->GraphPrepend(prevVertice);
    else
        tail_vertice = prevVertice;
    // (TBD) remove "vertice" from all referenced Queues                      
}  // end ProtoGraph::RemoveVertice()



ProtoGraph::VerticeIterator::VerticeIterator(const ProtoGraph& theGraph)
  : graph(theGraph)
{
    Reset();
}

ProtoGraph::VerticeIterator::~VerticeIterator()
{
}

void ProtoGraph::VerticeIterator::Reset()
{
    next_vertice = graph.head_vertice;
}

ProtoGraph::SimpleTraversal::SimpleTraversal(const ProtoGraph& theGraph, 
                                 Vertice&          startVertice,
                                 bool              depthFirst)
 : graph(theGraph), start_vertice(startVertice), 
   depth_first(depthFirst), current_level(0), trans_vertice(NULL),
   queue_pending(&item_pool), queue_visited(&item_pool)
{
    Reset();
}

ProtoGraph::SimpleTraversal::~SimpleTraversal()
{
    queue_pending.Empty();
    queue_visited.Empty();
    item_pool.Destroy();
}

bool ProtoGraph::SimpleTraversal::Reset()
{
    queue_pending.Empty();
    queue_visited.Empty();
    // Put "start_vertice" at head of pending queue
    if (!queue_pending.Prepend(start_vertice))
    {
        DMSG(0, "ProtoGraph::SimpleTraversal::Reset() error: couldn't enqueue vertice\n");
        return false;
    }
    trans_vertice = &start_vertice;
    current_level = 0;
    return true;
}  // end ProtoGraph::SimpleTraversal::Reset()

ProtoGraph::Vertice* ProtoGraph::SimpleTraversal::GetNextVertice(unsigned int* level)
{
    Vertice* currentVertice = queue_pending.GetHead();
    if (NULL != currentVertice)
    {
        queue_pending.Transfer(*currentVertice, queue_visited);
        // Iterate through edges, enqueue "unvisited" vertices
        // for future visitation (breadth first search)
        ProtoGraph::Vertice::EdgeIterator edgeIterator(*currentVertice);
        Edge* nextEdge;
        Edge* firstEdge = NULL;
        while ((nextEdge = edgeIterator.GetNextEdge()))
        {
            Vertice& nextVertice = nextEdge->GetDst(*currentVertice);
            if (!nextVertice.IsInQueue(queue_pending) &&
                !nextVertice.IsInQueue(queue_visited))
            {
                // unvisited vertice
                if (depth_first)
                    queue_pending.Prepend(nextVertice);
                else
                    queue_pending.Append(nextVertice);
                if (NULL == firstEdge) firstEdge = nextEdge;
            }
        }
        // Track depth in search tree as search progresses ...
        if (NULL == trans_vertice)
        {
            trans_vertice = firstEdge ? &firstEdge->GetDst(*currentVertice) : NULL;
        }
        else if (trans_vertice == currentVertice)
        {
            trans_vertice = firstEdge ? &firstEdge->GetDst(*currentVertice) : NULL; 
            current_level++;  
        }
        if (NULL != level) *level = current_level;
        return currentVertice;
    }
}  // end ProtoGraph::SimpleTraversal::GetNextVertice()
