#include "manetGraph.h"
#include <protoDebug.h>



ManetGraph::Cost::Cost()
{
   
}

ManetGraph::Cost::~Cost()
{
}


bool ManetGraph::Cost::operator>(const Cost& cost) const
{
    if (cost.IsMax())
    {
        return false;
    }
    else if (IsMax())
    {
        return true;
    }
    else if (UseSignBit())
    {
        bool mySign = (0x80 & *GetKeyValue());
        bool urSign = (0x80 & *cost.GetKeyValue());
        if (mySign)
        {
            if (urSign)
            {
                unsigned int keyLength = GetKeyLength();
                if (cost.GetKeyLength() < keyLength) keyLength = cost.GetKeyLength();
                if (UseComplement2())
                    return (0 < memcmp(GetKeyValue(), cost.GetKeyValue(), keyLength));
                else
                    return (0 > memcmp(GetKeyValue(), cost.GetKeyValue(), keyLength));
            }
            else
            {
                return false;
            }
        }
        else if (urSign)
        {
            return true;
        }
        else
        {
            unsigned int keyLength = GetKeyLength();
            if (cost.GetKeyLength() < keyLength) keyLength = cost.GetKeyLength();
            return (0 < memcmp(GetKeyValue(), cost.GetKeyValue(), keyLength));
        }
    }
    else
    {
        unsigned int keyLength = GetKeyLength();
        if (cost.GetKeyLength() < keyLength) keyLength = cost.GetKeyLength();
        return (0 < memcmp(GetKeyValue(), cost.GetKeyValue(), keyLength));
    }    
}  // end ManetGraph::Cost::operator>()


const double ManetGraph::SimpleCost::MAX_VALUE = ManetGraph::SimpleCost::ComputeMaxValue();
const ManetGraph::SimpleCost ManetGraph::COST_MAX(false);

ManetGraph::SimpleCost::SimpleCost(bool minimize)
{
     if (minimize)
        Minimize();
    else
        Maximize();
}

ManetGraph::SimpleCost::~SimpleCost()
{
}

ManetGraph::SimpleCost::SimpleCost(const SimpleCost& cost)
 : value(cost.value)
{
}

bool ManetGraph::SimpleCost::Copy(const Cost& cost)
{
    const SimpleCost* simpleCost = dynamic_cast<const SimpleCost*>(&cost);
    if (NULL != simpleCost)
    {
        value = simpleCost->value;
        return true;
    }
    else
    {
        DMSG(0, "ManetGraph::SimpleCost::Copy() error: different cost types\n");
        return false;
    }   
}  // end ManetGraph::SimpleCost::Copy()

double ManetGraph::SimpleCost::ComputeMaxValue()
{
    return htond(-1.0);
}  // end ManetGraph::SimpleCost::ComputeMaxValue()

void ManetGraph::SimpleCost::operator+=(const SimpleCost& cost)
{
    if (IsMax())
        return;
    else if (cost.IsMax())
         Maximize();
    else
        value = htond(ntohd(value) + ntohd(cost.value));
}  // end ManetGraph::SimpleCost::operator+=()


ManetGraph::Link::Link(Interface& a, Interface& b)
 : Edge(a, b), cost_a(NULL), cost_b(NULL)
{  
}

ManetGraph::Link::~Link()
{   
    if ((NULL != cost_b) && (cost_b != cost_a))
        delete cost_b;
    if (NULL != cost_a)
        delete cost_a;
    cost_a = cost_b = NULL;
}

bool ManetGraph::Link::SetCost(const Interface& src, const Cost& theCost)
{
    ASSERT((&src == &term_a) || (&src == &term_b));
    // To which direction is the cost being applied
    Cost** myCostHandle = (&src == &term_a) ? &cost_a : & cost_b;
    return SetCostInternal(myCostHandle, theCost);
}  // end ManetGraph::Link::SetCost()


bool ManetGraph::Link::SetCost(const Cost& theCost)
{
    // Apply same cost to both directions of the link
    Cost** myCostHandle = (NULL != cost_a) ? &cost_a : & cost_b;
    Cost** urCostHandle = (NULL != cost_a) ? &cost_b : & cost_a;
    
    bool wasAsymmetric = (cost_a != cost_b);
    
    if (SetCostInternal(myCostHandle, theCost))
    {
        if (wasAsymmetric && (NULL != *urCostHandle))
            delete *urCostHandle;
        *urCostHandle = *myCostHandle;
        return true;
    }
    else
    {
        return false;
    }
}  // end ManetGraph::Link::SetCost()


bool ManetGraph::Link::SetCostInternal(Cost** costHandle, const Cost& theCost)
{
    if (NULL != *costHandle)
    {
        if (!((*costHandle)->Copy(theCost)))
        {
            Cost* tmpCostPtr = theCost.Clone();
            if (NULL != tmpCostPtr)
            {
               DMSG(0, "ManetGraph::Link::SetCostInternal() clone error: %d\n",
                       GetErrorString());
               return false;   
            }
            if (cost_b != cost_a)
                delete *costHandle;
            *costHandle = tmpCostPtr; 
        }
    }
    else
    {
        *costHandle = theCost.Clone();
        if (NULL == *costHandle)
        {
           DMSG(0, "ManetGraph::Link::SetCostInternal() clone error: %d\n",
                   GetErrorString());
           return false;   
        }
    }
    return true;
}  // end ManetGraph::Link::SetCostInternal()

        
ManetGraph::Interface::Interface(Node& theNode, const ProtoAddress& addr)
 : node(theNode), address(addr), node_next(NULL)
{
    ProtoTree::Item::Init(address.GetRawHostAddress(), address.GetLength() << 3);
}

ManetGraph::Interface::~Interface()
{   
    
} 

bool ManetGraph::Interface::Connect(Interface& iface, const Cost& linkCost, bool duplex)
{ 
    // 1) Does the link already exist?
    LinkIterator linkIterator(*this);
    Link* link;
    while (NULL != (link = linkIterator.GetNextLink()))
    {
        if (&iface == &link->GetDst(*this))
            break;
    }
    // 2) If not, create it
    if (NULL == link)
    {
        if (NULL == (link = new Link(*this, iface)))
        {
            DMSG(0, "ManetGraph::Interface::Connect() new Link error: %s\n",
                    GetErrorString());
            return false;
        }
    }
    // 3) Set the link cost from "this" -> "iface"
    if (!link->SetCost(*this, linkCost))
    {
        DMSG(0, "ManetGraph::Interface::Connect() error setting link cost!\n");
        Disconnect(iface, duplex);
        return false;
    }
    
    // 4) Make it bi-directional if "duplex"
    if (duplex)
    {
        if (!link->GetDst(*this).Connect(*this, linkCost, false))
        {
            DMSG(0, "ManetGraph::Interface::Connect() error setting duplex link\n");
            Disconnect(iface, duplex);
            return false;
        }
    }
    return true;
}  // end ManetGraph::Interface::Link()


void ManetGraph::Interface::Disconnect(Interface& iface, bool duplex)
{
    // 1) Does the link exist?
    LinkIterator linkIterator(*this);
    Link* link;
    while (NULL != (link = linkIterator.GetNextLink()))
    {
        if (&iface == &link->GetDst(*this))
            break;
    }
    if (NULL != link)
    {
        if (duplex)
        {
            //link->DetachFromVertices();
            delete link;
        }
        else
        {
            ASSERT(0);  // (TBD) implement simplex disconnect
        }
    }
}  // end ManetGraph::Interface::Disconnect()

ManetGraph::Interface::LinkIterator::LinkIterator(Interface& theInterface)
 : EdgeIterator(static_cast<Vertice&>(theInterface))
{
}

ManetGraph::Interface::LinkIterator::~LinkIterator()
{
}


ManetGraph::Interface::PriorityQueue::Item::Item()
{
}

ManetGraph::Interface::PriorityQueue::Item::~Item()
{
    // (TBD) delete cost ...
}

bool ManetGraph::Interface::PriorityQueue::Item::SetCost(const Cost& theCost)
{
    // Do we already have a "cost" value set?
    if (NULL != cost)
    {
        if (!(cost->Copy(theCost)))
        {
            Cost* newCost = theCost.Clone();
            if (NULL != newCost)
            {
               DMSG(0, "ManetGraph::Interface::PriorityQueue::Item::SetCost() clone error: %d\n",
                       GetErrorString());
               return false;   
            }
            delete cost;
            cost = newCost; 
            // Make our ProtoSortedTree::Item() refer to the new cost as key for sorting
            TRACE("Initing cost keysize to %d\n", cost->GetKeyLength() << 3);
            ProtoSortedTree::Item::Init(cost->GetKeyValue(), cost->GetKeyLength());
        }
    }
    else
    {
        cost = theCost.Clone();
        if (NULL == cost)
        {
           DMSG(0, "ManetGraph::Interface::PriorityQueue::Item::SetCost() clone error: %d\n",
                   GetErrorString());
           return false;   
        }
        // Make our ProtoSortedTree::Item() refer to the new cost as key for sorting
        TRACE("Initing cost keysize to %d\n", cost->GetKeyLength() << 3);
        ProtoSortedTree::Item::Init(cost->GetKeyValue(), cost->GetKeyLength());
    }
    return true;
}  // end ManetGraph::Interface::PriorityQueue::Item::SetCost()

ManetGraph::Interface::PriorityQueue::PriorityQueue(PriorityQueue::ItemPool* itemPool)
 : item_pool(itemPool)
{
}

ManetGraph::Interface::PriorityQueue::~PriorityQueue()
{
    TRACE("priority queue dtor ...\n");
    Empty();
}


bool ManetGraph::Interface::PriorityQueue::Insert(Interface& iface, const Cost& cost)
{
    Item* item = GetNewItem();
    if (NULL == item)
    {
        DMSG(0, "ManetGraph::Interface::PriorityQueue::Insert() error: couldn't allocate item\n");
        return false;
    }
    TRACE("Setting cost for item:%p iface:%s\n", item, iface.GetAddress().GetHostString());
    item->SetQueue(this);
    Associate(*item, iface);
    item->SetCost(cost);
    InsertItem(*item);
    return true;
}  // end ManetGraph::Interface::PriorityQueue::Insert()

bool ManetGraph::Interface::PriorityQueue::Prepend(Interface& iface)
{
    Item* item = GetNewItem();
    if (NULL == item)
    {
        DMSG(0, "ManetGraph::Interface::PriorityQueue::Prepend() error: couldn't allocate item\n");
        return false;
    }
    Associate(*item, iface);
    PrependItem(*item);
    return true;
}  // end ManetGraph::Interface::PriorityQueue::Prepend()

bool ManetGraph::Interface::PriorityQueue::Append(Interface& iface)
{
    Item* item = GetNewItem();
    if (NULL == item)
    {
        DMSG(0, "ManetGraph::Interface::PriorityQueue::Append() error: couldn't allocate item\n");
        return false;
    }
    Associate(*item, iface);
    AppendItem(*item);
    return true;
}  // end ManetGraph::Interface::PriorityQueue::Append()


void ManetGraph::Interface::PriorityQueue::Remove(Interface& iface)
{
    Item* item = static_cast<Item*>(GetQueueState(iface));
    ASSERT(NULL != item);
    RemoveItem(*item);
    Disassociate(*item, iface);
    if (NULL != item_pool)
        item_pool->Put(*item);
    else
        delete item;
}  // end ManetGraph::Interface::PriorityQueue::Remove()

ManetGraph::Interface* ManetGraph::Interface::PriorityQueue::RemoveHead()
{
    TRACE("enter ManetGraph::Interface::PriorityQueue::RemoveHead() ...\n");
    Item* item = static_cast<Item*>(iface_list.GetHead());
    TRACE("   got head %p ....\n", item);
    if (NULL != item)
    {
        Interface* iface = item->GetInterface();
        ASSERT(NULL != iface);
        TRACE("   iface = %s\n", iface->GetAddress().GetHostString());
        
        //const SimpleCost* cost = static_cast<const SimpleCost*>(item->GetCost());
        //ASSERT(cost != NULL);
        //TRACE("   removing head item (cost:%lf) \n", cost->GetValue());
        RemoveItem(*item);
        
        TRACE("   disassociating ...\n", item);
        Disassociate(*item, *iface);
        TRACE("   pooling ...\n", item);
        if (NULL != item_pool)
            item_pool->Put(*item);
        else
            delete item;
        return iface;
    }
    else
    {
        return NULL;
    }
}  // end ManetGraph::Interface::PriorityQueue::RemoveHead()

void ManetGraph::Interface::PriorityQueue::Adjust(Interface& iface, const Cost& newCost)
{
    Item* item = static_cast<Item*>(GetQueueState(iface));
    ASSERT(item != NULL);
    iface_list.Remove(*item);
    item->SetCost(newCost);
    iface_list.Insert(*item);
}  // end ProtoGraph::Vertice::PriorityQueue::Adjust()

bool ManetGraph::Interface::PriorityQueue::AdjustDownward(Interface& iface, const Cost& newCost)
{
    Item* item = static_cast<Item*>(GetQueueState(iface));
    ASSERT(item != NULL);
    const Cost* itemCost = item->GetCost();
    ASSERT(NULL != itemCost);
    if (newCost < *itemCost)
    {
        iface_list.Remove(*item);
        item->SetCost(newCost);
        iface_list.Insert(*item);
        return true;
    }
    else
    {
        return false;
    }
}  // end ProtoGraph::Vertice::PriorityQueue::AdjustDownward()


void ManetGraph::Interface::PriorityQueue::Empty()
{
    while (!IsEmpty()) 
    {
        TRACE("in ManetGraph::Interface::PriorityQueue::Empty() loop\n");
        RemoveHead();
    }
    TRACE("exit empty\n");
}  // end ManetGraph::Interface::PriorityQueue::Empty()


void ManetGraph::Interface::PriorityQueue::InsertItem(PriorityQueue::Item& item)
{
    ASSERT(NULL != item.GetCost());
    if (!iface_list.IsReady())  
    {
        // Init our ProtoTree 
        Interface* iface = item.GetInterface();
        ASSERT(NULL != iface);
        unsigned int costBytes = item.GetCost()->GetKeyLength();// << 3;
        bool useSignBit = item.GetCost()->UseSignBit();
        bool useComplement2 = item.GetCost()->UseComplement2();
        iface_list.Init(costBytes, useSignBit, useComplement2);
    } 
    TRACE("INSERTING item with keysize:%d\n", item.GetKeysize());
    iface_list.Insert(item);
}  // end ManetGraph::Interface::PriorityQueue::InsertItem()

ManetGraph::Interface* ManetGraph::Interface::PriorityQueue::GetNextHop(const Interface& dstIface, const Interface& srcIface) const
{
    Item* item = static_cast<Item*>(GetQueueState(dstIface));
    if (NULL != item)
    {
        Link* nextHopLink = item->GetNextHopLink();
        if (NULL != nextHopLink)
            return (&nextHopLink->GetDst(srcIface));
        else
            return NULL;
    }
    else
    {
        return NULL;  // likely not connected
    }
}  // end ManetGraph::Interface::PriorityQueue::GetNextHop()

ManetGraph::Interface* ManetGraph::Interface::PriorityQueue::GetPrevHop(const Interface& dstIface) const
{
    Item* item = static_cast<Item*>(GetQueueState(dstIface));
    if (NULL != item)
    {
        Link* parentLink = item->GetParentLink();
        return ((NULL != parentLink) ? &parentLink->GetDst(dstIface) : NULL);
    }
    else
    {
        return NULL;  // likely not connected
    }
}  // end ManetGraph::Interface::PriorityQueue::GetNextHop()



ManetGraph::Interface::PriorityQueue::ItemPool::ItemPool()
 : head(NULL)
{
}

ManetGraph::Interface::PriorityQueue::ItemPool::~ItemPool()
{
    Destroy();
}

void ManetGraph::Interface::PriorityQueue::ItemPool::Destroy()
{
    while (!IsEmpty())
    {
        Item* item  = Get();
        delete item;
    }   
}  // end ManetGraph::Interface::PriorityQueue::ItemPool::Destroy()

ManetGraph::Interface::PriorityQueue::Item* ManetGraph::Interface::PriorityQueue::ItemPool::Get()
{
    if (IsEmpty())
    {
        Item* item = new Item;
        if (NULL == item)
            DMSG(0, "ManetGraph::Interface::PriorityQueue::ItemPool::Get() new Item error: %s\n",
                    GetErrorString());
        return item;
    }
    else
    {
        Item* item = head;
        head = static_cast<Item*>(head->GetPoolNext());
        return item;
    }
}  // end ManetGraph::Interface::PriorityQueue::ItemPool::Get()

// tbd make the pqueue iterator a member of this instead???
ManetGraph::Interface::PriorityQueue::Iterator::Iterator(PriorityQueue& theQueue, bool reverse)
 : ProtoSortedTree::Iterator(theQueue.iface_list)
{
}

ManetGraph::Interface::PriorityQueue::Iterator::~Iterator()
{
}


ManetGraph::Node::Node(const ProtoAddress& defaultAddr)
 : default_interface(*this, defaultAddr)
{
}

ManetGraph::Node::~Node()
{
    // Delete any extra interfaces
    Interface* next = default_interface.GetNodeNext();
    while (NULL != next)
    {
        Interface* iface = next;
        Interface* next = next->GetNodeNext();
        delete iface;
    }
}       

void ManetGraph::Node::AppendInterface(Interface& iface)
{
    iface.Append(NULL);
    Interface* prev = NULL;
    Interface* next = &default_interface;
    while (next)
    {
        prev = next;
        next = next->GetNodeNext();   
    }
    prev->Append(&iface);
}  // end ManetGraph::Node::AppendInterface()

void ManetGraph::Node::RemoveInterface(Interface& iface)
{
    ASSERT(&iface != &default_interface);
    Interface* prev = NULL;
    Interface* next = &default_interface;
    while (next)
    {
        if (&iface == &default_interface)
        {
            DMSG(0, "ManetGraph::Node::RemoveInterface() error: can't remove default_interface\n");
        }
        else if (&iface == next)
        {
            prev->Append(iface.GetNodeNext());
            return;
        }
        prev = next;
        next = next->GetNodeNext();   
    }
}  // end ManetGraph::Node::RemoveInterface()

ManetGraph::ManetGraph()
{
}

ManetGraph::~ManetGraph()
{
}

void ManetGraph::InsertNode(Node& node, Interface* iface)
{
    // need to init iface_tree on first insert !!!
    if (NULL == iface) iface = &node.GetDefaultInterface();
    iface_tree.Insert(*iface);
    ProtoGraph::InsertVertice(*iface);
}  // end ManetGraph::InsertNode()

void ManetGraph::RemoveNode(Node& node, Interface* iface)
{
    if (NULL == iface)
    {
        // Remove all interfaces associated with this node
        iface = &node.GetDefaultInterface();
        while (NULL != iface)
        {
            RemoveVertice(static_cast<Vertice&>(*iface));
            iface_tree.Remove(*iface);
            iface = iface->GetNodeNext();
        }
    }
    else
    {   
        // Remove only the interface specified
        iface_tree.Remove(*iface);
        RemoveVertice(static_cast<Vertice&>(*iface));
    }                                  
}  // end ManetGraph::RemoveNode()


ManetGraph::InterfaceIterator::InterfaceIterator(ManetGraph& theGraph)
  : ProtoGraph::VerticeIterator(static_cast<ProtoGraph&>(theGraph))
{
}

ManetGraph::InterfaceIterator::~InterfaceIterator()
{
}

ManetGraph::DijkstraTraversal::DijkstraTraversal(ManetGraph& theGraph, 
                                                 Node&       startNode,
                                                 Interface*  startIface)
 : manet_graph(theGraph),
   start_iface((NULL != startIface) ? *startIface : startNode.GetDefaultInterface()),
   trans_iface(NULL), current_level(0), dijkstra_completed(false), in_update(false)
{
    ASSERT(&start_iface.GetNode() == &startNode);
    Reset();
}

ManetGraph::DijkstraTraversal::~DijkstraTraversal()
{
    TRACE("enter ManetGraph::DijkstraTraversal::~DijkstraTraversal() ...\n");
    queue_pending.Empty();
    TRACE("emptying queue_visited ....\n");
    queue_visited.Empty();
    TRACE("exit ManetGraph::DijkstraTraversal::~DijkstraTraversal().n");
    //item_pool.Destroy();
}

bool ManetGraph::DijkstraTraversal::Reset()
{
    // (TBD) use a dual queue approach so that we can avoid 
    //      visiting every iface twice everytime we run the Dijkstra.
    //      I.e., we could maintain "visited" & "unvisited" queue, removing ifaces
    //      from that queue as they are visited, and then at the end of
    //      the Dijkstra, update the cost of any remaining unvisited ifaces
    //      to COST_MAX and then swap visited and unvisited queues for next
    //      Dijkstra run ... Thus each iface would be visited only once _or_ as
    //      needed for Dijkstra instead of once _plus_ as needed for Dijkstra
    queue_visited.Empty();
    queue_pending.Empty();
    SimpleCost startCost;
    startCost.Minimize();
    if (!queue_pending.Insert(start_iface, startCost))
    {
        DMSG(0, "ManetGraph::DijkstraTraversal::Reset() error: couldn't enqueue start_iface\n");
        return false;
    }
    dijkstra_completed = false;
}  // end ManetGraph::DijkstraTraversal::Reset()


// Dijkstra traversal step
ManetGraph::Interface* ManetGraph::DijkstraTraversal::GetNextInterface()
{
    Interface* currentIface = queue_pending.GetHead();
    if (NULL != currentIface)
    {
        queue_pending.Transfer(*currentIface, queue_visited);
        const SimpleCost* currentCost = static_cast<const SimpleCost*>(queue_visited.GetCost(*currentIface));
        ASSERT(NULL != currentCost);
        TRACE("cost to node %s = %lf\n", currentIface->GetAddress().GetHostString(), currentCost->GetValue());
        Interface::LinkIterator linkIterator(*currentIface);
        Link* nextLink;
        while ((nextLink = linkIterator.GetNextLink()))
        {
            // (TBD) we could do something different here,
            //       like have a virtual comparator method
            //       so this Dijkstra would work on derived
            //       variations of the "Cost", "Link", and 
            //       "Interface" classes
            Interface& nextDst = nextLink->GetDst(*currentIface);
            const SimpleCost* linkCost = static_cast<const SimpleCost*>(nextLink->GetCost(*currentIface));
            ASSERT(NULL != linkCost);
            TRACE("   link to %s cost = %lf\n", nextDst.GetAddress().GetHostString(), linkCost->GetValue());
            SimpleCost nextCost = static_cast<const SimpleCost&>(*linkCost);
            nextCost += *currentCost;
            TRACE("     cost sum = %lf\n", nextCost.GetValue());
            
            
            if (nextDst.IsInQueue(queue_pending))
            {
                // We have found another path to this pending iface.
                // If it is a shorter path update the cost to lower value
                if (!queue_pending.AdjustDownward(nextDst, nextCost))
                        continue;
            }
            else if (!nextDst.IsInQueue(queue_visited))
            {
                // This is the first path found to this iface,
                // so enqueue it in our "queue_pending"
                if (!queue_pending.Insert(nextDst, nextCost))
                {
                    DMSG(0, "ManetGraph::DijkstraTraversal::GetNextInterface() error: couldn't enqueue iface\n");
                    return NULL;
                }
            }
            else if (in_update)
            {
                // Check for shorter path to a "visited" iface
                // (must be in "Update()" mode!)
                if (queue_visited.AdjustDownward(nextDst, nextCost))
                {
                    // Update the routing tree state 
                    if (currentIface == &start_iface)
                        queue_visited.SetRouteLinks(nextDst, nextLink, nextLink);
                    else
                        queue_visited.SetRouteLinks(nextDst, queue_visited.GetNextHopLink(*currentIface), nextLink);
                }
                continue;
            }
            else
            {
                continue;
            }
            
            // print out queue pending
            Interface::PriorityQueue::Iterator it(queue_pending);
            Interface* i;
            while (NULL != (i = it.GetNextInterface()))
            {
                const SimpleCost* c = static_cast<const SimpleCost*>(queue_pending.GetCost(*i));
                ASSERT(NULL != c);
                TRACE("queue_pending i: %s cost:%lf\n", i->GetAddress().GetHostString(), c->GetValue());
            }
            TRACE("\n");
                    
                    
            // Save our routing tree state as we go
            if (currentIface == &start_iface)
                queue_pending.SetRouteLinks(nextDst, nextLink, nextLink);
            else
                queue_pending.SetRouteLinks(nextDst, queue_visited.GetNextHopLink(*currentIface), nextLink);
        }  // end while ((nextLink = linkIterator.GetNextLink()))
        
    }
    else
    {
        dijkstra_completed = true;
    }   
    return currentIface;
}  // end ManetGraph::DijkstraTraversal::GetNextInterface()

/*
// Dijkstra traversal step
ManetGraph::Interface* ManetGraph::DijkstraTraversal::GetNextInterface()
{
    Interface* currentIface = static_cast<Interface*>(queue.RemoveHead());
    if (NULL != currentIface)
    {
        const Cost& currentCost = currentIface->GetCost();
        ASSERT(!currentCost.IsMax());
        Interface::LinkIterator linkIterator(*currentIface);
        Link* nextLink;
        while ((nextLink = linkIterator.GetNextLink()))
        {
            nextLink->Unselect(*currentIface);
            // (TBD) we could do something different here,
            //       like have a virtual comparator method
            //       so this Dijkstra would work on derived
            //       variations of the "Cost", "Link", and 
            //       "Interface" classes
            Cost nextCost = nextLink->GetCost(*currentIface);
            nextCost += currentCost;
            Interface& nextDst = nextLink->GetDst(*currentIface);
            if (nextCost < nextDst.GetCost()) 
            {
                if (nextDst.GetCost().IsMax())
                {
                   // This is the first path found
                    nextDst.SetCost(nextCost);
                    queue.InsertFromTail(nextDst); 
                }
                else
                {
                    // We found a shorter path 
                    // (never called when links are of homogeneous cost)
                    queue.Adjust(nextDst, nextCost);
                }

                // Update shortest path info
                nextDst.SetParentLink(nextLink);
                if (currentIface == &start_iface)
                    nextDst.SetNextHopLink(nextLink);
                else
                    nextDst.SetNextHopLink(currentIface->GetNextHopLink());
            }
        }
        Link* parentLink = currentIface->GetParentLink();
        if (NULL != parentLink)
            parentLink->Select(parentLink->GetDst(*currentIface));
    }
    else
    {
        dijkstra_completed = true;
    }
    return currentIface;
}  // end ManetGraph::DijkstraTraversal::GetNextInterface()
*/
        
void ManetGraph::DijkstraTraversal::Update(Interface& startIface)
{
    in_update = true;
    // (TBD) fix this for new Dijkstra approach
    if (!dijkstra_completed)
    {
        // Complete dijkstra traversal
        while (NULL != GetNextInterface());
    }
    
    ASSERT(queue_pending.IsEmpty());
    
    if (startIface.IsInQueue(queue_visited))
    {
        queue_visited.Transfer(startIface, queue_pending);
    }
    else
    {
        ASSERT(0);
    }
    
    while (NULL != GetNextInterface());
    in_update = false;
}  // end ManetGraph::DijkstraTraversal::Update()

// Call this to setup re-traverse of tree computed via Dijkstra
bool ManetGraph::DijkstraTraversal::TreeWalkReset()
{
    // If Dijkstra was not completed, run full Dijkstra
    if (!dijkstra_completed)
    {
        Reset();
        while (NULL != GetNextInterface());
    }
    ASSERT(queue_pending.IsEmpty());
    if (!queue_pending.Append(start_iface))
    {
        DMSG(0, "ManetGraph::DijkstraTraversal::TreeWalkReset() error: couldn't append start_iface\n");
        return false;
    }
    trans_iface = NULL;
    current_level = 0;
}  // end MrbNodeTree::Traversal::TreeWalkReset()

ManetGraph::Interface* ManetGraph::DijkstraTraversal::TreeWalkNext(unsigned int* level)
{
    Interface* currentIface =  queue_pending.RemoveHead();
    if (NULL != currentIface)
    {
        // Find selected links
        Interface::LinkIterator linkIterator(*currentIface);
        Link* nextLink;
        Link* firstLink = NULL;
        while ((nextLink = linkIterator.GetNextLink()))
        {
            const Interface& nextDst = nextLink->GetDst(*currentIface);
            if (currentIface == GetPrevHop(nextDst))
            {
                if (NULL == firstLink) firstLink = nextLink;
                queue_pending.Append(nextLink->GetDst(*currentIface));
            }
        }
        // Track depth as walk progresses ...
        if (NULL == trans_iface)
        {
            trans_iface = firstLink ? &firstLink->GetDst(*currentIface) : NULL;
        }
        else if (trans_iface == currentIface)
        {
            trans_iface = firstLink ? &firstLink->GetDst(*currentIface) : NULL; 
            current_level++;  
        }
        if (NULL != level) *level = current_level;
    }
    return currentIface;
}  // end ManetGraph::DijkstraTraversal::TreeWalkNext()
 
