/*
 *  TCPProtoSocketAgent.cpp
 *
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */

#include "nsTCPProtoSocketAgent.h"
   
/////////////////////////////////////////////////////////////////////////
// NsProtoSimAgent::UdpSocketAgent implementation
//
static class ProtoSocketTcpAgentClass : public TclClass {
public:
	ProtoSocketTcpAgentClass() : TclClass("Agent/ProtoSocket/TCP") {}
	TclObject* create(int, const char*const*) {
		return (new NsTCPProtoSocketAgent());
	}
} class_proto_socket_tcp_agent;

NsTCPProtoSocketAgent::NsTCPProtoSocketAgent()
 : Agent(PT_TCP)
{
	socketType_=NotInitialised;
	txBufferSize=65536; // default buffers
	rxBufferSize=65536; 
//	printf("NsTCPProtoSocketAgent: Created ok\n");
}
 
bool NsTCPProtoSocketAgent::Listen(UINT16 thePort) {
	if (socketType_==NotInitialised) { // must be a server socket, port is set from the NsProtoSimAgent bind operation

		printf("TCPProtoSocketAgent: Listening on %i\n", thePort);
		
		Agent* simAgent = dynamic_cast<Agent*>(proto_socket->GetNotifier());

		serverSocket = TCPSocketFactory::createServerSocketAgent(simAgent);
		serverSocket->setTCPAgentProtocol(TCPSocketAgent::FULLTCP);
		
		serverSocket->bind(thePort);		
		socketType_=TcpServerSocket;
		
//		printf("TCPProtoSocketAgent: Calling listen on server socket\n");
		
		// start listening for connections
		serverSocket->listen();
		
		serverSocket->setListener(this);

//		printf("TCPProtoSocketAgent: Server initialsed and listening\n");

		}
	else {
		return false;
	}
 
	return true;
	}

bool NsTCPProtoSocketAgent::Connect(const ProtoAddress& theAddress) { 
//	printf("TCPProtoSocketAgent: Connecting\n");
	
	if (socketType_==NotInitialised) { // must be a client socket, port is set from the NsProtoSimAgent bind operation		
		Agent* simAgent = dynamic_cast<Agent*>(proto_socket->GetNotifier());
		tcpPipe = TCPSocketFactory::createSocketAgent(simAgent); // sets simAgent as agent to bind to
		tcpPipe->setTCPAgentProtocol(TCPSocketAgent::FULLTCP);

		tcpPipe->setTxBufferSize(txBufferSize);
		tcpPipe->setRxBufferSize(rxBufferSize);

		tcpPipe->bind(port());
		tcpPipe->setListener(this); // notifications sent here

		socketType_=TcpSocket;
		
		// connect to server
		tcpPipe->connect(atoi(theAddress.GetRawHostAddress()), theAddress.GetPort());
		}
	else {
		return false;
	}
			 
	return true; 
	} 

 /** 
  * Here, we create a new protosocket to
  * deal with the new connection and tell this TCPProtoSocketAgent to reply to the requesting
  * client. This allows us to spwan off new agents to deal with multiple clients.
  * NOTE that this assumes that the user provides a socket to deal with the connection !
  */ 
bool NsTCPProtoSocketAgent::Accept(ProtoSocket* theSocket) { 
//	printf("TCPProtoSocketAgent::Accept CALLED \n");
	
	theSocket->SetNotifier(proto_socket->GetNotifier());

    // Note that the first argument does not matter because the third is false i.e. bindOnCreate is false
	// so port is not used in Open
	theSocket->Open(0, ProtoAddress::SIM, FALSE); //creates a new nsTCPProtoSocketAgent
 
	NsTCPProtoSocketAgent *theProtoSocket = (NsTCPProtoSocketAgent *) static_cast<ProtoSimAgent::SocketProxy*>(theSocket->GetHandle());
	
	TCPSocketAgent* socket = serverSocket->accept(); // create new socket agent for this connection
	socket->setTxBufferSize(txBufferSize);
	socket->setRxBufferSize(rxBufferSize);

 	socket->setListener(theProtoSocket); // set the listener for RECEIVE events to this class 
	
	theProtoSocket->setTCPSocketAgent(socket);  // set the socket agent in the new TCPProtoSocketAgent to be 
	
	return true;
	} 
 

bool NsTCPProtoSocketAgent::SendTo(const char* buffer, unsigned int& numBytes, const ProtoAddress& dstAddr) {
	tcpPipe->send(numBytes,buffer); 
	return true;
}	
 
/**
 * Catches events thrown from the underlying TCP agents. 
 */ 
bool NsTCPProtoSocketAgent::tcpEventReceived(TCPEvent *event) {

	// check how these map ...

	printf("NsTCPPrototSocketAgent: Getting Callback\n");
	
	if (event->getType()==TCPEvent::ACCEPT) {
		if (proto_socket)
			proto_socket->OnNotify(ProtoSocket::NOTIFY_INPUT);  
	} else if (event->getType()==TCPEvent::CONNECTED) {
	//	printf("Connected event in ProtoSocketAgent\n");
		if (proto_socket)
			proto_socket->OnNotify(ProtoSocket::NOTIFY_OUTPUT);  
	} else if (event->getType()==TCPEvent::SEND) {
		if (proto_socket)
			proto_socket->OnNotify(ProtoSocket::NOTIFY_OUTPUT);  
	} else if (event->getType()==TCPEvent::RECEIVE) {
		recv_data = (char *)event->getData();
		recv_data_len = event->getDataSize();

		printf("NsTCPPrototSocketAgent: Got data of length %i = %s\n", recv_data_len, recv_data);
	//	printf("Destination port and Address are %i and %i\n", tcpPipe->getDestinationPort(), 
	//						tcpPipe->getDestinationAddress()); 

		src_addr.SimSetAddress(tcpPipe->getDestinationAddress());
		src_addr.SetPort(tcpPipe->getDestinationPort());
 
		if (proto_socket) 
			proto_socket->OnNotify(ProtoSocket::NOTIFY_INPUT);
				
	} else if (event->getType()==TCPEvent::DISCONNECTED) {
		if (proto_socket)
			proto_socket->OnNotify(ProtoSocket::NOTIFY_OUTPUT);  
	} else {
		printf("TCPProtoSocketAgent UNKNOWN Event on node %i\n", addr());
		return false;
		}
	
	printf("NsTCPPrototSocketAgent: finished Callback\n");

	delete event;
	return true;
} 
