/*
 *  TCPServerSocketAgent.cpp
 *
 *  Created by Ian Taylor on 01/03/2007.
 *
 */
 
#include "TCPServerSocketAgent.h" 

static class TCPServerSocketAgentClass : public TclClass {
public:
	TCPServerSocketAgentClass() : TclClass("Agent/TCP/ServerSocketAgent") {}
	TclObject* create(int argc, const char*const *argv) {		
		cout << "Creating TCPSocketAgent " << argc << " arguments" << endl;
		
		if (argc == 6) { // with protocol included
			return (new TCPServerSocketAgent(argv[4], atoi(argv[5])));
		} else if (argc == 4) {
			TCPServerSocketAgent *tcpsock = new TCPServerSocketAgent();
			return (tcpsock);
		} else
			return NULL;
	}
} class_tcp_server_socket_agent; 


void TCPServerSocketAgent::recv(Packet *pkt, Handler *handler) {
		printf("TCPServerSocketAgent:: recv Called \n");
		int tiflags = hdr_tcp::access(pkt)->flags() ; 		 // tcp flags from packet
		hdr_ip *iph = hdr_ip::access(pkt);

		if (tiflags & TH_SYN) { // connection request
			synPacket_= pkt;
			handler_=handler;
			// set the senders port and address
			dport() = iph->sport();
	 		daddr() = iph->saddr();  
			
	 		printf("TCPServerSocketAgent received a SYN - connection request from %i, port %i\n", iph->saddr(), iph->sport());

			if (tcpListener_!=NULL)
				tcpListener_->tcpEventReceived(new TCPEvent(TCPEvent::ACCEPT, this, NULL, NULL));
			else
				accept(); // no listener, so accept by default.
		} else { // its data to go to a TCPSocketAgent

			TCPSocketAgent* socket = mySocketList->findSocketByDestination(iph->saddr(), iph->sport());
			if (socket==NULL) {
				cout << "TCPServerSocketAgent: No socket found for address " << iph->saddr() << " and port " << iph->sport() << " !!!" << endl;
				cout << "TCPServerSocketAgent: This should not happen - please report bug " << endl;
				exit(1);
			}
			else { // its data so pass it to TCP agent
				printf("TCPServerSocketAgent:: Demultiplexing - passing onto TCP agent to receive data \n");
				
				// If this is the same node, just get the data and bypass TCP
				
				if (socket->getAddress()==socket->getDestinationAddress())
					socket->getTCPApp()->receiveDataAndNotify();
				else
					socket->getTCPAgent()->recv(pkt,handler);
				printf("TCPServerSocketAgent:: Packet Handled ok \n");
				}
			}
	}

 /**
  * Returns a bew TCPSocketAgent that will deal with the accepted connection
  */
TCPSocketAgent* TCPServerSocketAgent::accept() { 
	printf("TCPServerSocketAgent::Accept CALLED - we are on node %i, accepting connection from node %i, port %i\n", addr(), daddr(), dport());
	
	Tcl& tcl = Tcl::instance();    
 
	tcl.evalf("eval new Agent/TCP/SocketAgent");
	
	const char *tcpVar = tcl.result();
 
	if (tcpVar==NULL) return NULL;
		
	TCPSocketAgent *socket = (TCPSocketAgent *)tcl.lookup(tcpVar);
	socket->setSimulationAgent(initialSimulationAgent);
	
//	printf("TCPServerSocketAgent::Created new TCPSocketAgent \n");
 	  
	socket->setTCPAgentProtocol(theTCPProtocol_); // creates the underlying TCP protocol for the agent
		
	socket->setDestinationPort(dport()); 
	socket->setDestinationAddress(daddr());

  	socket->attachTCPAgentToNode(this, -1); // attaches the socket to this node, on a port negative 1 - unlikely to be taken ...
	
	socket->getTCPInterface()->listenOn(port()); // se we can receive data

	TCPSocketAgent *remoteSocket=NULL; 

    if (socket->getAddress() == socket->getDestinationAddress()) {
		remoteSocket = TCPSocketAgent::getSocketConnections()->findSocketByDestination(socket->getAddress(), socket->getPort()); // adds it to the static list so that connections can be set up
		TCPSocketAgent::getSocketConnections()->remove(remoteSocket);
		}

	TCPSocketAgent::getSocketConnections()->prepend(socket);
		
//	cout << "TCPServerSocketAgent: adding to socket list socket with address " << socket->getDestinationAddress() << " and port " << socket->getDestinationPort() << endl;

	mySocketList->prepend(socket);

//	printf("TCPServerSocketAgent::Forwarding SYN to Agent \n");

    // here only pass it to the TCP agent if we are not on the same node. If we are the TCPAgent breaks
	// so we bypass the protocol.
	
    if (socket->getAddress() != socket->getDestinationAddress())
		socket->getTCPAgent()->recv(synPacket_, handler_); // send the SYN packet to the TCP protocol so it can respond
														// as if it received it in the first place ....	 	
	else // need to notify connected socket...
		remoteSocket->tcpEventReceived(new TCPEvent(TCPEvent::CONNECTED, NULL, NULL, 0));

	return socket;
	}

bool TCPServerSocketAgent::bind(nsaddr_t thePort) { 
	printf("TCPServerSocketAgent: Attempting to Attach on port %i \n", thePort);

	if (initialSimulationAgent!=NULL) {
		attachTCPServerToNode(initialSimulationAgent, thePort); 
		printf("TCPServerSocketAgent: Attach Complete, TCP ServerAgent on node %i, port %i \n", addr(), port());
		return TRUE;
		}
	else {
		printf("TCPServerSocketAgent: Can't attach server to node: no simulation agent set\n");
		return false;
		}
	}

/**
 * Atatches the TCP agent to the node that the provided agent is attached to. The provide agent must have
 * been attached before this method is called, otherwise we cannot get a reference to the node. 
 */
bool TCPServerSocketAgent::attachTCPServerToNode(Agent *myAgent, nsaddr_t thePort) {
	Tcl& tcl = Tcl::instance();    
     
	tcl.evalf("%s set node_", myAgent->name());
	
	const char* nodeName = tcl.result();

	attachTCPServerToNode(nodeName, thePort);
	return true;
} 
	
/**
 * Overridden to become a dummy TCPSocketAgent that muli-plexes messages to real socket Agents.
 */
bool TCPServerSocketAgent::attachTCPServerToNode(const char *nodeNameInTCL, nsaddr_t thePort) {
    
	Tcl& tcl = Tcl::instance();	
    // find out the name of this node in tcl space
	
	// tcl.evalf("%s set node_", node->name());
    // const char* nodeName = tcl.result();

	// cout << "Node name for " << name() << " on node: " << nodeNameInTCL_ <<  endl;

//    if (thePort > 0)
        tcl.evalf("%s attach %s %hu", nodeNameInTCL, name(), thePort);
//    else
  //      tcl.evalf("%s attach %s", nodeNameInTCL, name());
		    		
	return false;  
	}
		
int TCPServerSocketAgent::command(int argc, const char*const* argv) {

	cout << "Node = " << name() << ", command: " << argv[1] << "," ;

    for (int i =0; i<argc; ++i)
		cout << " arg[" << i << "] = " << argv[i] << ", ";

	cout << endl;

	if (argc==2) {
		if (strcmp(argv[1], "listen") == 0) {
			//
			return (TCL_OK);
			}
	} else if (argc==3) {
		if (strcmp(argv[1], "setProtocol") == 0) {
			theTCPProtocol_=TCPSocketAgent::getTCPProtocolFromString(argv[2]); 
			return (TCL_OK);
		} else if (strcmp(argv[1], "attach-to-node") == 0) {
			attachTCPServerToNode(argv[2], 0); 
			return (TCL_OK);
		}
	}
	
	return Agent::command(argc, argv);
} 

