/*
 *  TCPSocketAgent.cpp
 *
 *  Created by Ian Taylor on 01/01/2007.
 *
 */
 

#include "TCPSocketAgent.h"
  
// These are the  overloaded TCP classes that just detect the sequence of events when 
// connecting - more intelligent TCPFull classes. They are NOT instantiated directly for use
// with TCPSocketAgent. Use TCPSocketAgent or TCPProtoSocketAgent instead 
  
  
// NOTE: we same name as original TCP for FullTcp so we get initialised correctly. Otherwise
// we do not get the MANY variables deinfed in ./tcl/lib/ns-default.tcl initialised correctly
static class FullTcpWithEventsClass : public TclClass { 
public:
	FullTcpWithEventsClass() : TclClass("Agent/TCP/FullTcp/Events") {} 
	TclObject* create(int argc, const char*const* argv) {  
		if (argc != 4)
			return NULL;
		return (new TCPFullWithEvents());
	}
} class_full_tcp_with_events;
 
static class TCPSocketAgentClass : public TclClass {
public:
	TCPSocketAgentClass() : TclClass("Agent/TCP/SocketAgent") {}
	TclObject* create(int argc, const char*const *argv) {				
		if (argc == 6) { // with protocol included
			return (new TCPSocketAgent(argv[4], atoi(argv[5])));
		} else if (argc == 4) {
			TCPSocketAgent *tcpsock = new TCPSocketAgent();
			return (tcpsock);
		} else
			return NULL;
	}  
} class_tcp_socket_agent; 


// Constructor Method Implementations

TCPSocketAgent::TCPSocketAgent(const char *theTCPProtocolString, nsaddr_t thePort)  : Agent(PT_TCP) {
	init();
	setTCPAgentProtocol(getTCPProtocolFromString(theTCPProtocolString));
	setPort(thePort);
}

TCPSocketAgent::TCPSocketAgent(TcpProtocol theTCPProtocol, nsaddr_t thePort) : Agent(PT_TCP) { 
	init(); 
	setTCPAgentProtocol(theTCPProtocol); 
	setPort(thePort); 
	}


TCPSocketAgent::TCPSocketAgent(FullTcpAgent *theTCPAgent, nsaddr_t thePort) : Agent(PT_TCP), tcpAgent(theTCPAgent) {
	init(); 
	tcpAgent=theTCPAgent;
	setPort(thePort);
	}

TCPSocketAgent::TcpProtocol TCPSocketAgent::getTCPProtocolFromString(const char *theTCPProtocolString) { 			
	if (strcmp(theTCPProtocolString, "FULLTCP") == 0)
		return FULLTCP;
	else if (strcmp(theTCPProtocolString, "RENOFULLTCP") == 0)
		return RENOFULLTCP;
	else if (strcmp(theTCPProtocolString, "SACKFULLTCP") == 0)
		return SACKFULLTCP;
	else if (strcmp(theTCPProtocolString, "TAHOEFULLTCP") == 0)
		return TAHOEFULLTCP;
	else 
		return ERROR;
	}
  

const char *TCPSocketAgent::getTCPStringForProtocol(TCPSocketAgent::TcpProtocol theTCPProtocol) {
	if (theTCPProtocol == FULLTCP)
		return "FULLTCP";
	else if (theTCPProtocol == RENOFULLTCP)
		return "RENOFULLTCP";
	else if (theTCPProtocol == SACKFULLTCP)
		return "SACKFULLTCP";
	else if (theTCPProtocol == TAHOEFULLTCP)
		return "TAHOEFULLTCP";
	else 
		return "ERROR";
}

SocketList *TCPSocketAgent::getSocketConnections() { 
	static SocketList *socketConns = new SocketList(); 
	return socketConns; 
 	}
 
bool TCPSocketAgent::bind(nsaddr_t thePort) { 	
 	printf("TCPSocketAgent::bind - Attaching on port %hu\n", thePort);

		if (initialSimulationAgent!=NULL) {
			attachTCPAgentToNode(initialSimulationAgent, thePort);
			printf("TCPSocketAgent: Attach Complete: TCP Agent on node %i, port %i \n", getAddress(), getPort());
			return TRUE;
			}
		else { 
			printf("TCPSocketAgent: Can't attach socket to node: no simulation agent set\n");
			return false;
			}
	}


bool TCPSocketAgent::connect(TCPSocketAgent *remoteTCPSockAgent) { 			
	FullTcpAgent *remoteFullTCPAgent = remoteTCPSockAgent->getTCPAgent();	
	connect(remoteFullTCPAgent->addr(), remoteFullTCPAgent->port());
	
	return true;
} 

/**
  Connect to remote location.  Note this just sets the dst port
*/
bool TCPSocketAgent::connect(nsaddr_t destAddress, nsaddr_t destPort) {
	printf("TCPSocketAgent: Node %i, connecting to node address %i, port %i\n", getAddress(), destAddress, destPort);

	tcpAgentInterface->setDestinationPort(destPort);
	tcpAgentInterface->setDestinationAddress(destAddress);  
	
	if (destAddress==getAddress()) // add it to the list so we can make a connection on local nodes
		TCPSocketAgent::getSocketConnections()->prepend(this);
	
	connect();

	printf("TCPSocketAgent: Node %i connected to address %i, port %i\n", getAddress(), destAddress, destPort);
 
	return true;
}	

	
/**
 * Send data to the TCP server through this socket. This uses the TCPSocket app class to actually
 * transfer the application data between the nodes.
 */	
void TCPSocketAgent::send(int nbytes, const char *data) {
	TcpData *tmp = new TcpData();

	std::cout << "TCPSocketAgent sending: " << (char *)(data +4) << endl;

	tmp->setBytes(data, nbytes);

	printf("TCPSocketAgent: I am sending from Node %i, port %i\n", getAddress(), getPort());
	printf("TCPSocketAgent: Sending %s to Node %i, port %i\n", data, getDestinationAddress(), getDestinationPort());
	
	tcpSocketApp->send(nbytes, tmp);
	
	printf("TCPSocketAgent: ok Sent\n");
} 

// TCPSocketAgentListener Interface implementation - used to receive events from TCP sockets

/**
 * Events from TCP Application received here - this can be reimplemented in a subclass in order to link into
 * whichever the software toolkit or application requires. Conenction events and receive etc defined in TCPEvent
 * are triggered by the underlying TCP agents and channeled through this method for conveneince. 
 */
bool TCPSocketAgent::tcpEventReceived(TCPEvent *event) {
//	printf("TCPSocketAgent: EVENT\n");
 
	if (event->getType()==TCPEvent::CONNECTED) {
		    printf("TCPSocketAgent CONNECTED Event on node %i\n", getTCPAgent()->addr());
	 					
			// Just set when SYN was received by remote node - find it in the list
			// we just look it up by using our own port and address as an id

			TCPSocketAgent* remoteTCPSocketAgent = 
	 			TCPSocketAgent::getSocketConnections()->findSocketByDestination(getAddress(),getPort()); 
			
			if (remoteTCPSocketAgent!=NULL) {
				TCPSocketApp *remoteTCPSocketApp = remoteTCPSocketAgent->getTCPApp();
 				remoteTCPSocketApp->connect(tcpSocketApp); // connect the apps together 
				TCPSocketAgent::getSocketConnections()->remove(remoteTCPSocketAgent); // remove it, connection is complete 
				}
			else {
				printf("FATAL: Could not retrieve remoteSocketAgent for connection on %i\n", getTCPAgent()->addr());
				exit(0);
				}
	} else if (event->getType()==TCPEvent::SEND) {
		 printf("TCPSocketAgent SEND Event on node %i\n", getTCPAgent()->addr());
	} else if (event->getType()==TCPEvent::RECEIVE) {
		printf("TCPSocketAgent RECEIVE Event on node %i\n", getTCPAgent()->addr());
		printf("TCPSocketAgent: Receiving %s of size %i at node %i\n", (char *)event->getData(), event->getDataSize(), getTCPAgent()->addr());
		printf("TCPSocketAgent: Done \n");
		}
	else if (event->getType()==TCPEvent::DISCONNECTED) 
 		printf("TCPSocketAgent DISCONNECT Event on node %i\n", getTCPAgent()->addr());
	else {
		printf("TCPSocketAgent UNKNOWN Event on node %i\n", getTCPAgent()->addr());
		return false;
		}
		 
	printf("TCPSocketAgent: Passing it onto listeners \n");

	assert(tcpListener_);
	
	// If a listener exists pass on the event
	if (tcpListener_) {
		event->setSourceObject(this); // state it came from this object, so application can keep track
		tcpListener_->tcpEventReceived(event);
	}

	printf("TCPSocketAgent: finished Callback\n");
	
	return true;
}
 
// Private Method Implementations
 
void TCPSocketAgent::setTCPParameter(const char *parName, const char *parValue) {
	Tcl& tcl = Tcl::instance();    
	
//	cout << "Setting parameter " << parName << " to " << parValue << " on agent " << tcpAgent->name() << " on node: " << tcpAgent->addr() << endl;

	tcl.evalf("%s set %s %s", tcpAgent->name(), parName, parValue);
}
 
/**
 * Attaches the tcpAgent i.e. a FullTcp variant to this Ns-2 node.
 */
bool TCPSocketAgent::attachTCPAgentToNode(const char *nodeNameInTCL, nsaddr_t thePort) {

	Tcl& tcl = Tcl::instance();    

//	cout << "Attaching agent = " << tcpAgent->name() << " on node: " << nodeNameInTCL <<  endl;
 

   if (thePort >= 0) {
    	printf("TCPSocketAgent::attachTCPAgentToNode - Attaching on port %hu\n", thePort);
        tcl.evalf("%s attach %s %hu", nodeNameInTCL, tcpAgent->name(), thePort);
		}
    else {
    	printf("TCPSocketAgent::attachTCPAgentToNode - Attaching on ANY available port -1 was specified\n");	    
        tcl.evalf("%s attach %s", nodeNameInTCL, tcpAgent->name());
	}
	
	return false; 
	}
 
/**
 * Atatches the TCP agent to the node that the provided agent is attached to. The provide agent must have
 * been attached before this method is called, otherwise we cannot get a reference to the node. 
 */
bool TCPSocketAgent::attachTCPAgentToNode(Agent *myAgent, nsaddr_t thePort) {
	Tcl& tcl = Tcl::instance();    
     
	tcl.evalf("%s set node_", myAgent->name());
	
	const char* nodeName = tcl.result();

	attachTCPAgentToNode(nodeName, thePort);
	return true;
} 

bool TCPSocketAgent::setTCPAgentProtocol(TcpProtocol theTCPProtocol){
	Tcl& tcl = Tcl::instance();    
	
	theTCPProtocol_=theTCPProtocol;
	
	switch (theTCPProtocol) {
		case FULLTCP :  tcl.evalf("eval new Agent/TCP/FullTcp/Events");
			cout << "Created FULLTCP Agent" << endl;
			break;
		case TAHOEFULLTCP : tcl.evalf("eval new Agent/TCP/FullTcp/Tahoe");
			cout << "Created FULLTCP Tahoe Agent" << endl;
			break;
		case RENOFULLTCP : tcl.evalf("eval new Agent/TCP/FullTcp/Newreno");
			cout << "Created FULLTCP Newreno Agent" << endl;
			break;
		case SACKFULLTCP : tcl.evalf("eval new Agent/TCP/FullTcp/Sack");
			cout << "Created FULLTCP Sack Agent" << endl;
			break;
		default : protocolError("TCPSocketAgent: No TCP Identifier Specified");
	}
	
	const char *tcpVar = tcl.result();
	 
    tcpAgent = (FullTcpAgent *)tcl.lookup(tcpVar);

	// cout << "tcpAgent Name " << tcpAgent->name() << endl;

    if (tcpAgent==NULL) {
		cerr << "TCPSocketAgent: Cannot instantiate TCP Agent " << tcpVar << endl;
		exit(1);
		}
	    	
	createTCPApplication(tcpAgent);
		
	// hook up the callback mechanism from the TCP agents to the Apps
	
	switch (theTCPProtocol) {
		case FULLTCP : ((TCPFullWithEvents *)tcpAgent)->setTCPSocketAgent(this);
			tcpAgentInterface = (TCPAgentInterface *)((TCPFullWithEvents *)tcpAgent);
	 		break;
		case TAHOEFULLTCP : 
			break;
		case RENOFULLTCP : 
			break;
		case SACKFULLTCP : 
			break;
		case ERROR : 
			return false;
	}
	
	return true;
	}

void TCPSocketAgent::protocolError(const char *theTCPProtocolString) {
	cerr << "TCPSocketAgent: Unable to identify TCP Protocol String" << theTCPProtocolString << endl;
	cerr << "TCPSocketAgent: Allowed Types are " << endl;
	cerr << "FULLTCP, RENOFULLTCP, SACKFULLTCP, TAHOEFULLTCP" << endl;
	exit(1);
}

/**
 * Creates the TCP application for sending data between TCP nodes. 
 */
bool TCPSocketAgent::createTCPApplication(FullTcpAgent *theTCPAgent) {
	
	Tcl& tcl = Tcl::instance();    

	tcl.evalf("eval new Application/TCPSocketApp %s", theTCPAgent->name());
 	
	const char *tcpVar = tcl.result();
	
    tcpSocketApp = (TCPSocketApp *)tcl.lookup(tcpVar);

	// Create Socket app and binds the TCPSocketApp with the underlying TCP protocol

	// cout << "Created Socket App " << tcpSocketApp->name() << " on node " << theTCPAgent->addr() << endl;
		
	tcpSocketApp->setTCPSocketAgentListener(this); // add ourselves as a listener for events - i.e. we get notified
 
	tcpSocketApp->setTxBufferSize(txBufferSize);
	tcpSocketApp->setRxBufferSize(rxBufferSize);
	
	return true;
    }
 	
	
int TCPSocketAgent::command(int argc, const char*const* argv) {
	Tcl& tcl = Tcl::instance();    
//	cout << "Node = " << name() << ", command: " << argv[1] << "," ;
 
//    for (int i =0; i<argc; ++i)
//		cout << " arg[" << i << "] = " << argv[i] << ", ";

	cout << endl; 

	if (argc==2) {
		if (strcmp(argv[1], "listen") == 0) {
			listen(port());
			return (TCL_OK);
			}
	} else if (argc==3) {	
		if (strcmp(argv[1], "setProtocol") == 0) {
			setTCPAgentProtocol(getTCPProtocolFromString(argv[2])); 
			return (TCL_OK);
		} else if (strcmp(argv[1], "attach-to-node") == 0) {
			attachTCPAgentToNode(argv[2],0); 
			return (TCL_OK);
		} else if (strcmp(argv[1], "tcp-connect") == 0) {
			TCPSocketAgent *tcpSocket = (TCPSocketAgent *)TclObject::lookup(argv[2]);
			if (tcpSocket == NULL) {
				tcl.resultf("%s: connected to null object.", name_);
				return (TCL_ERROR);
			}
			connect(tcpSocket);
 		
			return (TCL_OK);
		}
	} else if (argc==4) {
		// Pass all the set commands to the TCL agent e.g. $tcp1 set window_ 100
		if (strcmp(argv[1], "set-var") == 0) { 
			setTCPParameter(argv[2], argv[3]);
		
			return (TCL_OK);
		} else if (strcmp(argv[1], "tcp-connect") == 0) {
		//	cout << "Trying to connect " << argv[2] << " port " << argv[2] << endl;
			nsaddr_t address = atoi(argv[2]);
			nsaddr_t port = atoi(argv[3]);
			connect(address,port);
			return (TCL_OK);
		} else if (strcmp(argv[1], "send") == 0) {
			const char *bytes = argv[3];
			int size = atoi(argv[2]);
		
//			cout << "Sending " << bytes << ", size " << size << " from node " << getTCPAgent()->addr() << endl;
		
			send(size,bytes); // the socketapp actually sends the num bytes to the protocol through our previous app 
							  // attachment layer. We just have to pass the data across.
		
			return (TCL_OK);
		}
	}
	
	return Agent::command(argc, argv);
}    

SocketList::SocketList() : head(NULL) {
}

void SocketList::prepend(TCPSocketAgent *proxy) {
//	printf("SocketList: Prepend \n");
	SocketListNode *listnode = new SocketListNode();
	 
	listnode->setSocket(proxy); // wrap object so we can have multiple lists
	
    listnode->setPrev(NULL);
    listnode->setNext(head);
    if (head) head->setPrev(listnode);
    head = listnode;
} 
 
void SocketList::remove(TCPSocketAgent *proxy) {
//	printf("SocketList: remove \n");
    SocketListNode* next = head;
	SocketListNode* proxyNode=NULL;
	
    while (next) {
        if (next->getSocket()==proxy) {
			proxyNode=next;
			break;
			}
        else
            next = next->getNext();   
		}

	if (proxyNode==NULL) {
		printf("WARNING !!!!! - can't remove item from list\n");
		return; 
		}
		
	SocketListNode* prev = proxyNode->getPrev();
    next = proxyNode->getNext();

    if (prev) 
        prev->setNext(next);
    else
        head = next;
    if (next)
        next->setPrev(prev);
	
	delete proxyNode; // delete container for socket
} 
 
TCPSocketAgent* SocketList::findSocketByDestination(nsaddr_t destinationAddress, nsaddr_t destinationPort) {					
    SocketListNode* next = head;
    while (next) {
		cout << "SockeList: comparing " << next->getSocket()->getDestinationAddress() << " and port " << next->getSocket()->getDestinationPort() << endl;
		cout << "SockeList: with " << destinationAddress << " and port " << destinationPort << endl;
        if ((next->getSocket()->getDestinationPort() == destinationPort) && (next->getSocket()->getDestinationAddress() == destinationAddress))
            return next->getSocket();
        else
            next = next->getNext();   
    }
    return NULL;
}

void SocketList::removeSockets() {					
    SocketListNode* cur = head;
	SocketListNode* next = cur->getNext();
    while (cur) {
		delete cur->getSocket();
		delete cur;
		cur = next; 
		next = cur->getNext();   
    }
}


