
//
// Tcp application interface: extension of the TCPApp code to do duplex connections to transmit 
// real application data and also allow multiple TCP connections
// 
// The single connections model underlying TCP connection as a 
// FIFO byte stream, use this to deliver user data
 
#include "agent.h"
#include "app.h"
#include "TCPSocketApp.h"

 
// TCPSocketApp that accept TCP agents as a parameter

static class TcpAppIntClass : public TclClass {
public:
	TcpAppIntClass() : TclClass("Application/TCPSocketApp") {}
	TclObject* create(int argc, const char*const* argv) {
		if (argc != 5)
			return NULL;
		Agent *tcp = (Agent *)TclObject::lookup(argv[4]);
		if (tcp == NULL) 
			return NULL;
		return (new TCPSocketApp(tcp));
	}
} class_TCPSocketApp; 

// Socket Apps that don't have any underlying TCP protocol defined - these are used
// in the dynamic scenario, where we generate these.

TCPSocketApp::TCPSocketApp(Agent *tcp) : 
	Application(), curdata_(0), curbytes_(0), bytesInSendBuffer_(0)
{
	setTCPAgent(tcp);
}



TCPSocketApp::~TCPSocketApp()
{
	// XXX Before we quit, let our agent know what we no longer exist
	// so that it won't give us a call later...
	agent_->attachApp(NULL);
}

void TCPSocketApp::setTCPAgent(Agent *tcp) {
//	cout << "TCPSocketApp: setting Agent to " << tcp->name() << " and attaching app " << endl;
	
	agent_ = tcp;
	agent_->attachApp(this);
} 


void TCPSocketApp::receiveDataAndNotify() {

		curdata_ = dst_->rcvr_retrieve_data();

		process_data(curdata_->size(), curdata_->data());
		// Then cleanup this data transmission
//		printf("TCPSocketApp: Deleting curdata_\n");
		delete curdata_;
}		

AppData* TCPSocketApp::get_data(int&, AppData*) {
	printf("TCPSokcetApp: get_data called - aborting ...\n");
		// Not supported
	abort();
	return NULL;
	}

/**   
 * Send with calls to add data to this pipe by adding to a FIFO buffer before allowing
 * the underlying TCP implementation to make a simulated transfer.
 */
void TCPSocketApp::send(int nbytes, AppData *cbk)
{
	printf("TCPSocketApp:: Send Called\n");

	TcpData *tmp = (TcpData*)cbk;

	printf("TCPSocketApp:: Send Called after cast\n");

//	printf("TCPSocketApp:: Send Called, sending %s bytes\n", tmp->getDataSize());
//	std::cout << "TCPSocketApp sending: " << (char *)(tmp->getBytes() +4) << endl;
	
	TcpDataFIFO *p = new TcpDataFIFO(cbk, nbytes);
#ifdef TCPSocketApp_DEBUG
	p->time() = Scheduler::instance().clock();
#endif
	TcpDataFIFO_.insert(p);

// send number of bytes to the underlying protocol i.e. FullTCP or whatever

	bytesInSendBuffer_+=nbytes;

	Application::send(nbytes);
	
	// notify the TCP socket agent with a SEND event (i.e. buffer is ready to write to as its empty now)
	
//	if (bytesInSendBuffer_ < txBufferSize) { // send a SEND event if there's room in the buffer
//		if (tcpSocketListenerAgent) 
//			tcpSocketListenerAgent->tcpEventReceived(new TCPEvent(TCPEvent::SEND, NULL, NULL, 0));
//		}

//	printf("TCPSocketApp:: Send Called \n"); 
}
 
TcpDataFIFO* TCPSocketApp::rcvr_retrieve_data() { 
	TcpDataFIFO* data = TcpDataFIFO_.detach();
	
	bytesInSendBuffer_-= data->bytes(); // decrease the amount in buffer

	// must be some room now so send and SEND event
	
// MUST Fix SEND Messages ...
	// if (tcpSocketListenerAgent) 
	//	tcpSocketListenerAgent->tcpEventReceived(new TCPEvent(TCPEvent::SEND, NULL, NULL, 0));
	
//	printf("TCPSocketApp: Ending\n");
	 
	return data; 	
	}

// All we need to know is that our sink has received one message
void TCPSocketApp::recv(int size) {		
	printf("TCPSocketApp:: Recv Called \n");
	
	if (dst_==NULL) {
		printf("Destination for SocketApp is NULL -> the setup is incorrect. Check the code\n");
		exit(0);
		}

	// If it's the start of a new transmission, grab info from dest, 
	// and execute callback
	if (curdata_ == 0)
		curdata_ = dst_->rcvr_retrieve_data();
			
	if (curdata_ == 0) {
		fprintf(stderr, "[%g] %s receives a packet but no callback!\n",
			Scheduler::instance().clock(), name_);
		return;
	}

	curbytes_ += size;

	if (curbytes_ == curdata_->bytes()) {
		// We've got exactly the data we want
		// If we've received all data, execute the callback
//		printf("TCPSocketApp: To Process Data ---------------------------------------------------\n");
		process_data(curdata_->size(), curdata_->data());
		// Then cleanup this data transmission
//		printf("TCPSocketApp: Deleting curdata_\n");
		delete curdata_;
//		printf("TCPSocketApp: finished deleting\n");
		curdata_ = NULL;
		curbytes_ = 0;
	} else if (curbytes_ > curdata_->bytes()) {
		// We've got more than we expected. Must contain other data.
		// Continue process callbacks until the unfinished callback
		while (curbytes_ >= curdata_->bytes()) {
			process_data(curdata_->size(), curdata_->data());
			curbytes_ -= curdata_->bytes();
#ifdef TCPSocketApp_DEBUG
			fprintf(stderr, 
				"[%g] %s gets data size %d(left %d)\n", 
				Scheduler::instance().clock(), 
				name(),
				curdata_->bytes(), curbytes_);
				//curdata_->data());
#endif
			delete curdata_;
			curdata_ = dst_->rcvr_retrieve_data();
			if (curdata_ != 0)
 				continue;
			if ((curdata_ == 0) && (curbytes_ > 0)) {
				fprintf(stderr, "[%g] %s gets extra data!\n",
					Scheduler::instance().clock(), name_);
				Tcl::instance().eval("[Simulator instance] flush-trace");
				abort();
			} else
 				// Get out of the look without doing a check
				break;
		}
	}
//	printf("TCPSocketApp:: Recv Done \n");
}

void TCPSocketApp::resume()
{
	// Do nothing
}

int TCPSocketApp::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();

//	cout << "Command: " << argv[1] << endl;

	if (strcmp(argv[1], "connect") == 0) {
		dst_ = (TCPSocketApp *)TclObject::lookup(argv[2]);
		if (dst_ == NULL) {
			tcl.resultf("%s: connected to null object.", name_);
			return (TCL_ERROR);
		}
		dst_->connect(this);
		return (TCL_OK);
	} else if (strcmp(argv[1], "send") == 0) {
		
		const char *bytes = argv[3];
		int size = atoi(argv[2]);
		
//		cout << "Sending " << bytes << ", size " << size << " from node " << getTCPAgent()->addr() << endl;
		
		if (argc == 3)
			send(size, NULL);
		else {
			TcpData *tmp = new TcpData();
			tmp->setBytes(bytes, size);
			send(size, tmp);
		}
		 
		return (TCL_OK);

	} else if (strcmp(argv[1], "dst") == 0) {
		tcl.resultf("%s", dst_->name());
		return TCL_OK;
	}
	return Application::command(argc, argv);
}
   
/**
 * Application Callback HERE:
 * 
 * process_data() is called when data is delivered from the TCP agent. This effectively is the callback 
 * an application gets from an agent when data arrives. Since TCPSocketApp uses the application
 * Ns2 interface, then this method will be called upon receipt of data the the Ns-2 node.   
 */
void TCPSocketApp::process_data(int size, AppData* data) 
{
	if (data == NULL)
		return;
		
	TcpData *tmp = (TcpData*)data;
    
	printf("TCPSocketApp:: Recv called, receiving size %i, %s \n", tmp->getDataSize(), tmp->getBytes());

	if (target()) {
//		printf("TCPSocketApp:: Sending to Target\n");
		send_data(size, data);
	} else if (tcpSocketListenerAgent) { // pass it along
//		printf("TCPSocketApp:: Sending to Listener\n");
		tcpSocketListenerAgent->tcpEventReceived(new TCPEvent(TCPEvent::RECEIVE, NULL, tmp->getBytes(), tmp->getDataSize()));
//		printf("TCPSocketApp: finished Callback\n");
	}
	else {
		cout << "TCPSocketApp: Receiving " << tmp->getBytes() << " of size " << tmp->getDataSize()  << " at node " << getTCPAgent()->addr() << endl;
	}
}
