/*
 *  TCPSocketExample.h
 *
 *  Created by Ian Taylor on 01/03/2007.
 *
 */

#include <agent.h>
#include <ip.h>

#include "TCPSocketAgent.h"
#include "TCPServerSocketAgent.h"

#include "TCPSocketFactory.h"

// C++ example ... work in progress
// need to solve locating the TCL reference for a node so we can bind to it 

class TCPSocketExample : public Agent, TCPSocketAgentListener {
public:
	int CLIENT_CONNECTIONS; 
	int SERVER_CONNECTIONS; 
	int PORT;

	TCPSocketAgent **clients;
	TCPServerSocketAgent *server;
	TCPSocketAgent **serversockets;
	
	int connectTo; // address that the server is on
	
	int sockno;

	TCPSocketExample() : Agent(PT_TCP) { CLIENT_CONNECTIONS=1; SERVER_CONNECTIONS=1; PORT=1; sockno=0; }
	
	~TCPSocketExample() {}
	
	void createServer() {
		server = TCPSocketFactory::createServerSocketAgent(this); // must set yoursef as the simulation agent
		
		if (server->bind(PORT)==FALSE) printf("Problem with Binding node to port\n");
 		 
		server->setTCPAgentProtocol(TCPSocketAgent::FULLTCP);
		server->listen();
		server->setListener(this);

		printf("TCPSocketExample: Created server on port %i\n", server->port()); 

		printf("TCPSocketExample: Creating Server Sockets\n"); 
		
		serversockets = new TCPSocketAgent *[SERVER_CONNECTIONS];
		
		printf("TCPSocketExample: Created Server, ok\n"); 
	}
 
	void createClients() {
		clients = new TCPSocketAgent *[CLIENT_CONNECTIONS];
		
		for (int i=0; i<CLIENT_CONNECTIONS; ++i) {
			clients[i] = TCPSocketFactory::createSocketAgent(this);
			clients[i]->setTCPAgentProtocol(TCPSocketAgent::FULLTCP);
			clients[i]->bind(PORT+i);
			clients[i]->setListener(this);
		}
	printf("TCPSocketExample: Created Clients\n"); 
	}
 	
	void connectClients() {
		for (int i=0; i<CLIENT_CONNECTIONS; ++i) 
			clients[i]->connect(connectTo, PORT);
		printf("TCPSocketExample: Connected Clients\n"); 
		}
		
	void sendData() {
		char *message;
		int len;
		
		for (int i=0; i<CLIENT_CONNECTIONS; ++i) { 
			message= new char[20]; // max buffer size
			sprintf(message, "Hello from %i, port %i", addr(), clients[i]->port());
			len= strlen(message) +1; // send actual length
			printf("TCPSocketExample: Sending %s\n", message); 
			clients[i]->send(len, (const char *)message);
			delete[] message;
		}
	}
	
	virtual bool tcpEventReceived(TCPEvent *event) {
		printf("TCPSocketExample: An Event !, Type %i \n", event->getType()); 

		if (event->getType()==TCPEvent::SEND) {
			printf("TCPSocketExample: SEND Received - buffer ready to write\n");
			return true;
		}
		else if (event->getSource()==server) {
			if (event->getType()==TCPEvent::ACCEPT) {
				printf("TCPSocketExample: ACCEPT Notification\n");
	 			serversockets[sockno] = server->accept(); // accept the connection and get the socket responsible for it
				serversockets[sockno]->setListener(this); // set the listener for RECEIVE events to this class 
	 			++sockno; 
				return true;
			}		
		} else 	if (event->getType()==TCPEvent::RECEIVE) {
			printf("TCPSocketExample: Received Data\n");
			printf("TCPSocketExample: Received data \"%s\" of size %i at node %i\n", 
						(char *)event->getData(), event->getDataSize(), this->addr());
			return true;
		} else	if (event->getType()==TCPEvent::CONNECTED) {
			printf("TCPSocketExample: CONECTED to Server\n");
			return true;
			}
 	
	return false;
	}
		
	int command(int argc, const char*const* argv) {
		cout << "Node = " << name() << ", command: " << argv[1] << "," ;

		for (int i =0; i<argc; ++i)
			cout << " arg[" << i << "] = " << argv[i] << ", ";

		cout << endl;

		if (argc==2) {
			if (strcmp(argv[1], "createClients") == 0) {
				createClients();
				return (TCL_OK);
			} else if (strcmp(argv[1], "createServer") == 0) {
				createServer();
				return (TCL_OK);
			} else if (strcmp(argv[1], "connectClients") == 0) {
				connectClients();
				return (TCL_OK);
			} else if (strcmp(argv[1], "sendData") == 0) {
				sendData();
				return (TCL_OK);
			}
		} else if (argc==3) {
			if (strcmp(argv[1], "setClientConnections") == 0) {
				CLIENT_CONNECTIONS = atoi(argv[2]); 
				return (TCL_OK);
			}
			else if (strcmp(argv[1], "setServerConnections") == 0) {
				SERVER_CONNECTIONS = atoi(argv[2]); 
				return (TCL_OK);
			}
			else if (strcmp(argv[1], "setConnectTo") == 0) {
				connectTo = atoi(argv[2]); 
				return (TCL_OK);
			}
		}
	
		return Agent::command(argc, argv);
	} 
};

static class TCPSocketExampleClass : public TclClass { 
public:
	TCPSocketExampleClass() : TclClass("Agent/TCP/TCPSocketExample") {} 
	TclObject* create(int argc, const char*const* argv) {  
		if (argc != 4)
			return NULL;
		return (new TCPSocketExample());
	}
} class_TCP_Socket_Example;

